/*
 * Copyright (C) 1998 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)ld-linux.c	1.4	99/02/25 SMI"

#include	<errno.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/socket.h>
#include	<netdb.h>
#include	"lxrun.h"
#include	"auxv.h"
/* Remove compiler warnings due to previous defitions of AT_UID and AT_GID
   (in /usr/include/sys/vnode.h on Solaris x86). */
#undef AT_UID
#undef AT_GID
#include	"linux/elf.h"
#include	"version.h"
#include	"devmap.h"

char	emulator[PATH_MAX];	/* Contains full pathname of invoked lxrun */

extern int run_elf_interpreter(const char *name, char **argv, char **envp, 
			       auxv_t *auxv);
extern void lxrun_init(void);


/*
 * LINUX emulator for x86 *IX
 */
#if	defined(SOLARIS)

extern caddr_t	current_brk;
extern caddr_t  minimum_brk;

extern void	trace_init(void);

int
lxrun(const char *name, caddr_t end, char **argv, char **envp, auxv_t *auxv)
{
	char	*pathenv = getenv("PATH");

	find_emulator(pathenv, emulator, "lxrun");

	trace_init();
	lxrun_init();

	current_brk = minimum_brk = end;

	return (run_elf_interpreter(name, argv, envp, auxv));
}

#else

int
main(int argc, char **argv, char **envp)
{
	char	**pp;
	auxv_t	*auxv;
	auxv_t	*a;
	int	i;
	Elf32_Phdr	*phdr;
	int		phnum;
	int		phent;

	/*
	 * find the aux vector
	 */
	pp = envp;
	while (*pp)
		++pp;
	auxv = (auxv_t *) (pp + 1);

	for (a = auxv; a->a_type != AT_NULL; a++) {
		switch (a->a_type) {
			case AT_PHDR:
				phdr = a->a_un.a_ptr;
				break;
			case AT_PHENT:
				phent = a->a_un.a_val;
				break;
			case AT_PHNUM:
				phnum = a->a_un.a_val;
				break;
			default:
				/*
				 * UnixWare puts some additional entries
				 * in the aux vector which conflict with
				 * the Linux dynamic linker so we just
				 * ignore them
				 */
				if (a->a_type > AT_ENTRY)
					a->a_type = AT_IGNORE;
				break;
		}
	}

	for (i = 0; i < phnum; i++) {
		extern caddr_t	minimum_brk;
		extern caddr_t	current_brk;

		if (phdr[i].p_type == PT_LOAD) {
			current_brk =
			minimum_brk = (caddr_t)(phdr[i].p_vaddr + phdr[i].p_memsz);
		}
	}

	lxrun_init();

	run_elf_interpreter("/lib/ld-linux.so.1", argv, envp, auxv);

	fatal(errno, "can't load /lib/ld-linux.so.1");

	return 1;
}
#endif
