/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)trace.c	1.9	99/02/18 SMI"

#include	<errno.h>
#include	<stdarg.h>
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<signal.h>
#include	<fcntl.h>
#include	<sys/time.h>
#include	<sys/reg.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/resource.h>

#include	"lxrun.h"
#include	"sysent.h"
#include	"linux/ipc.h"
#include	"linux/socket.h"
#include	"linux/sigcontext.h"
#include	"trace.h"

#define	ARG0	regs[EBX]
#define	ARG1	regs[ECX]
#define	ARG2	regs[EDX]
#define	ARG3	regs[ESI]
#define	ARG4	regs[EDI]

#define	SYS(scall, rtype, args)	{ #scall, 0, rtype, { args } },

struct trace_info syscall_info[] =
{
#include "syscalls.h"
};

struct trace_info socket_info[] =
{
	SYS(socketcall,		RET(int), ARGS1(int))
	SYS(socket,		RET(int), ARGS3(int, int, int))
	SYS(bind,		RET(int), ARGS3(int, ptr, int))
	SYS(connect,		RET(int), ARGS3(int, ptr, int))
	SYS(listen,		RET(int), ARGS2(int, int))
	SYS(accept,		RET(int), ARGS3(int, ptr, ptr))
	SYS(getsockname,	RET(int), ARGS3(int, ptr, ptr))
	SYS(getpeername,	RET(int), ARGS3(int, ptr, ptr))
	SYS(socketpair,		RET(int), ARGS4(int, int, int, ptr))
	SYS(send,		RET(int), ARGS4(int, ptr, int, int))
	SYS(sendto,		RET(int), ARGS6(int, ptr, int, int, ptr, int))
	SYS(recv,		RET(int), ARGS4(int, ptr, int, int))
	SYS(recvfrom,		RET(int), ARGS6(int, ptr, int, int, ptr, ptr))
	SYS(shutdown,		RET(int), ARGS2(int, int))
	SYS(setsockopt,		RET(int), ARGS5(int, int, int, ptr, int))
	SYS(getsockopt,		RET(int), ARGS5(int, int, int, ptr, ptr))
};

struct trace_info ipc_info[] =
{
	SYS(ipc,		RET(int), ARGS1(int))
	SYS(semop,		RET(int), ARGS3(int, ptr, int))
	SYS(semget,		RET(int), ARGS3(int, int, ptr))
	SYS(semctl,		RET(int), ARGS4(int, int, hex, ptr))
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(msgsnd,		RET(int), ARGS4(int, ptr, int, ptr))
	SYS(msgrcv,		RET(int), ARGS5(int, ptr, int, ptr, ptr))
	SYS(msgget,		RET(int), ARGS2(int, int))
	SYS(msgctl,		RET(int), ARGS3(int, int, ptr))
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(ipc,		RET(int), ARGS0())
	SYS(shmat,		RET(int), ARGS4(int, ptr, int, hex))
	SYS(shmdt,		RET(int), ARGS1(ptr))
	SYS(shmget,		RET(int), ARGS3(int, int, hex))
	SYS(shmctl,		RET(int), ARGS3(int, int, hex))
};

int	trace_flag = 0;
FILE	*trace_fp;
static int	pid = -1;
int	masterpid = -1;

void
trace_init(void)
{
	if (getenv("LINUX_TRACE") != NULL)
		trace_flag = 1;
}

/*
 * tprintf(format, ...)
 */
void
tprintf(
	char	*fmt,		/* message format string		*/
	...			/* optional arguments			*/
	)
{
	va_list	args;
	int	trace_fd;
	struct rlimit rlim;

	if (trace_flag == 0)
		return;

	if (trace_fp == NULL)
	{
		char trace_name[PATH_MAX];
		struct stat st;
		if(masterpid == -1)
			sprintf(trace_name, "/tmp/lxrun.%lu", getpid());
		else
			sprintf(trace_name, "/tmp/lxrun.%lu-%lu", masterpid,
						getpid());
		/*
		** For security, we check to make sure the file
		** doesn't already exist.  It's possible some knave
		** symlinked /etc/passwd to /tmp/lxrun.123 or something.
		*/
		if(lstat(trace_name, &st) != -1 || errno != ENOENT)
			{
			fprintf(stderr, "lxrun: Warning: File %s already exists.\n        -- no trace output will be produced.\n", trace_name);
			trace_fd = open("/dev/null", O_WRONLY);
			}
		else
			trace_fd = open(trace_name,
			    O_WRONLY | O_TRUNC | O_CREAT, 0666);
		/*
		 * Now be sneaky and try to dup the fd above the soft
		 * limit, but below the hard limit. This is to avoid breaking
		 * when programs go close all their open file descriptors,
		 * which would confuse stdio.
		 */
		if ((getrlimit(RLIMIT_NOFILE, &rlim) != -1) &&
		    (rlim.rlim_cur < rlim.rlim_max)) {
			rlim_t	oldsoft;
			int	newfd, tmpfd;

			oldsoft = rlim.rlim_cur;
			rlim.rlim_cur = rlim.rlim_max;
			if (setrlimit(RLIMIT_NOFILE, &rlim) != -1) {
				for (newfd = oldsoft + 1;
				    newfd < rlim.rlim_max; newfd++) {
					/* see if 'newfd' is already taken */
					tmpfd = dup(newfd);
					if (tmpfd != -1) {
						(void) close(tmpfd);
						continue;

					} else if (errno != EBADF) {
						continue;
					}
					if (dup2(trace_fd, newfd) != -1) {
						(void) close(trace_fd);
						trace_fd = newfd;
						break;
					}
				}
				/* restore old soft limit */
				rlim.rlim_cur = oldsoft;
				(void) setrlimit(RLIMIT_NOFILE, &rlim);
			}
		}
		fcntl(trace_fd, F_SETFD, FD_CLOEXEC);
		trace_fp = fdopen(trace_fd, "w");
		setbuf(trace_fp, NULL);
	}
	va_start(args, fmt);
	(void) vfprintf(trace_fp, fmt, args);
	va_end(args);
}

void
trace_syscall(int n, int *regs, int rval, int err)
{
	struct	sysent	*s;
	int	args[8];
	char	buf[256];
	char	*p;

	if (trace_flag == 0)
		return;

	args[0]	= ARG0;
	args[1]	= ARG1;
	args[2]	= ARG2;
	args[3]	= ARG3;
	args[4]	= ARG4;
	args[5]	= 0;
	args[6]	= 0;
	args[7]	= 0;

	if (pid == -1)
		pid = getpid();

	if (n == 2)	/* fork() */
	{
		int	newpid = getpid();
		if (newpid != pid)	/* If we're a child... */
			{
			/* Keep the original lxrun invocation's pid */
			if(masterpid == -1)
				masterpid = getppid();

			/* Start a new trace file */
			trace_fp = NULL;

			pid = newpid;
			}
	}

	if (n >= 0 && n < nsysent)
	{
		int	i;
		struct trace_info	*t;
		t = &syscall_info[n];
		sprintf(buf, "%s(", t->name);
		p = &buf[strlen(buf)];

		for (i = 0; i < 6 && t->argtypes[i] != 0;)
		{
			if (i > 0)
				strcat(buf, ", ");
			i += PRINT_ARG(t, buf, args, i);

		}
		strcat(buf, ")");
		if (rval == -1) {
			strcat(buf, " ");
			strcat(buf, strerror(err));
		}
		else {
			strcat(buf, " returned ");
			PRINT_RET(t, buf, rval);
		}
		tprintf("%s\n", buf);
	} else {
		tprintf("syscall(%d)\n", n);
	}
}

void
trace_signal(int sig, void *info, void *context)
{
	if (trace_flag == 0)
		return;

	tprintf("pid %d: signal(%d, %#X, %#X)\n", pid, sig, info,
		context);
}

struct stackframe
{
    struct stackframe	*fp;		/* next stack frame	*/
    unsigned char	*ret;		/* return address	*/
    unsigned		arg[1];		/* function arguments	*/
};


static void
prname(unsigned long addr)
{
    if (addr == 0)
	tprintf("????????");
    else
	tprintf("%8x", addr);
}

/*
 * prframe()	- try to print out the contents of a stackframe
 *
 *		   understands 4 types of call:
 *			   e8 xx xx xx xx	call	xx xx xx xx
 *			ff 15 xx xx xx xx	call	ds:[xx xx xx xx]
 *			         ff 55 xx	call	ss:[ebp + xx]
 *			ff 95 xx xx xx xx	call	ss:[ebp + xx xx xx xx]
 */
static void
prframe(struct stackframe *fp)
{
    unsigned char	*r;
    unsigned long	f;
    int			ac;
    int			i;

    r = fp->ret;

    if (r[-5] == 0xe8)
	f = (unsigned long)r + *(long *)(r - 4);
    else if (r[-6] == 0xff && r[-5] == 0x15)
	f = **(unsigned long **)(r - 4);
    else if (r[-6] == 0xff && r[-5] == 0x95)
	f = *(unsigned long *)((unsigned char *)fp->fp + *(long *)(r - 4));
    else if (r[-3] == 0xff && r[-2] == 0x55)
	f = *(unsigned long *)((unsigned char *)fp->fp + *(char *)(r - 1));
    else
	f = 0;
    
    prname(f);		/* print function name	*/
    tprintf("(");

    if (r[0] == 0x83 && r[1] == 0xc4)		/* add esp, imm8	*/
	ac = r[2] / 4;
    else if (r[0] == 0x81 && r[1] == 0xc4)	/* add esp, imm32	*/
	ac = *(unsigned long *)&r[2] / 4;
    else if (r[0] == 0x59)			/* pop ecx		*/
	for (ac = 0; r[ac] == 0x59; ac++)
	    ;
    else
	ac = 0;

    for (i = 0; i < ac; i++)
	tprintf("%x%s", fp->arg[i], (i == ac - 1) ? "" : ", ");

    tprintf(")\t called from ");
    prname((unsigned long)r);
    tprintf("\n");

}

static void
trace_stack(void *frame)
{
	struct stackframe	*fp	= frame;

	for (fp = fp->fp; fp->fp != NULL; fp = fp->fp)
		prframe(fp);

	tprintf("\n");
}

void
trace_reg_changes(unsigned long *regs, struct lx_sigcontext_struct *lx_context)
{
	if (trace_flag == 0)
		return;

#define CHECK(REG, reg) \
	if (regs[REG] != lx_context->reg) \
		tprintf(#REG " changed from %8x to %8x\n", \
		    regs[REG], lx_context->reg)

	CHECK(EDI, edi);
	CHECK(ESI, esi);
	CHECK(EBP, ebp);
	CHECK(ESP, esp);
	CHECK(EBX, ebx);
	CHECK(EDX, edx);
	CHECK(ECX, ecx);
	CHECK(EAX, eax);
	CHECK(EIP, eip);
	CHECK(EFL, eflags);
#undef CHECK
}

void
trace_unexpected_sigsegv(void **regs, uchar_t *eip)
{
	unsigned long *sp = regs[UESP];

	if (trace_flag == 0)
		return;

	tprintf("\nCaught signal 11 at EIP = %#x\n", eip);
	tprintf("ESP = %08x, EBP = %08x, ESI = %08x, EDI = %08x\n",
	    regs[UESP], regs[EBP], regs[ESI], regs[EDI]);
	tprintf("EAX = %08x, EBX = %08x, ECX = %08x, EDX = %08x\n",
	    regs[EAX], regs[EBX], regs[ECX], regs[EDX]);

	tprintf("Stack top: %08x %08x %08x %08x\n",
	    sp[-4], sp[-3], sp[-2], sp[-1]);

	tprintf("Raising SIGSTOP on pid %d for a debugger.\n",
	    getpid());
	kill(getpid(), SIGSTOP);
	trace_stack(regs[EBP]);
	exit(1);
}

int
enosys(const char *funcname)
{
	tprintf("*** UNIMPLEMENTED syscall %s called ***\n", funcname);
	errno = ENOSYS;
	return -1;
}

int
is_remappable(int sysnum, int argnum)
{
	return (syscall_info[sysnum].argtypes[argnum] == a_path);
}
